local D = LibStub:NewLibrary("DoomCore-1.1",1)
if not D then return end
local SharedMedia = LibStub( "LibSharedMedia-3.0" )
DoomCore = D


-----------
-- Utility
-----------

local function empty(val)
  return val == nil or val == "" or (type(val) == "table" and next(val) == nil)
end
D.empty = empty

local function say(...)
  local arguments = { ... }
  local output = {}
  for i=1,#arguments do
    output[i] = tostring(arguments[i])
  end
  return table.concat(output)
end
D.say = say
local function tPrint(table) for k,v in pairs(table) do print(k,v) end end
D.tPrint = tPrint

local function stringify(val)
  local t_concat, t_insert = table.concat, table.insert
  if type(val) == "table" then
    if empty(val) then return "{}"
    elseif #val > 0 then
      local output = {}
      for i=1,#val do
        output[i] = tostring(val[i])
      end
      return "{" .. t_concat(output, ", ") .. "}"
    else
      local output = {}
      for k,v in pairs(val) do
        t_insert(output, tostring(k) .. " = " .. tostring(v))
      end
      return "{" .. t_concat(output, ", ") .. "}"
    end
  end
  return tostring(val)
end
D.stringify = stringify

local function assertNotNil(...)
  local t_insert, t_concat, math_ceil = table.insert, table.concat, math.ceil
  local arguments = { ... }
  local failures = {}
  for i=1,#arguments do
    if arguments[i] == nil then
      t_insert(failures, say("Arg #",math_ceil(i/2)," is nil."))
    end
  end
  if #failures>0 then error(t_concat(failures," "),2) end
  return ...
end
D.assertNotNil = assertNotNil

local function assertType(...)
  local t_insert, t_concat, math_ceil = table.insert, table.concat, math.ceil
  local arguments = { ... }
  local failures = {}
  for i=1,#arguments,2 do
    local isType = type(arguments[i])
    local assertType = arguments[i+1]
    local assertTypeType = type(assertType)
    if assertTypeType == "string" then
      if isType ~= assertType then
        t_insert(failures,
          say("Arg #",math_ceil(i/2)," must be ",assertType,", is ",isType, ". (" .. stringify(arguments[i]) .. ")"))
      end
    elseif assertTypeType == "table"  then
      local match = false
      for _,subType in ipairs(assertType) do
        if isType == subType then
          match = true
          break
        end
      end
      if not match then
        t_insert(failures, say("Arg #",math_ceil(i/2)," must be ",t_concat(assertType,"/"),", is ",isType, "."))
      end
    end
  end
  if #failures>0 then error(t_concat(failures," "),tonumber(arguments[#arguments]) or 2) end
end
D.assertType = assertType

local function nilSort(val, valFunc, sortFunc)
  table.sort(val, function(a, b)
    if valFunc then
      if a ~= nil then a = valFunc(a) end
      if b ~= nil then b = valFunc(b) end
    end
    if a == nil and b == nil then
      return false
    end
    if a == nil then
      return false
    end
    if b == nil then
      return true
    end
    if sortFunc then return sortFunc(a,b)
    else return a < b
    end
  end)
end
D.nilSort = nilSort

local function capitalize(val)
  assertType(val,"string", 3)
  return val:sub(1, 1):upper() .. val:sub(2):lower()
end
D.capitalize = capitalize

local function reverse(l)
  local t_insert = table.insert
  local newL = {}
  for i = getn(l), 1, -1 do t_insert(newL, l[i]) end
  return newL
end
D.reverse = reverse

local function flip(flip, ...)
  if flip then
    return unpack(reverse{ ... }) else
    return ...
  end
end
D.flip = flip

local function roundString(val)
  assertType(val,"number", 3)
  if val > 1000000 then
    return math.floor(val / 10000) / 100 .. "m"
  else
    return math.floor(val / 1000) .. "k"
  end
end
D.roundString = roundString

local function orderNum(text)
  local string_byte, math_pow = string.byte, math.pow
  assertType(text,"string", 3)
  text = text:upper()
  local order = 0
  for i=1,#text do
    order = order + (string_byte(text, i)) / math_pow(100, i)
  end
  return order
end
D.orderNum = orderNum

local function hexToCol(code)
  if empty(code) then return { r = 1, g = 1, b = 1, a = 1 } end
  return {
    r = tonumber("0x"..code:sub(1,2)) / 255,
    g = tonumber("0x"..code:sub(3,4)) / 255,
    b = tonumber("0x"..code:sub(5,6)) / 255,
    a = 1
  }
end
D.hexToCol = hexToCol

local function colToHue(col)
  assertType(col,"table", 3)
  local col_r, col_g, col_b = assertNotNil(col.r, col.g, col.b)
  local hue
  local min = math.min(col_r, col_g, col_b)
  local max = math.max(col_r, col_g, col_b)
  local range = max - min
  if max == col_r then
    hue = (col_g - col_b) / range
  elseif max == col_g then
    hue = 2 + (col_b - col_r) / range
  else
    hue = 4 + (col_r - col_g) / range
  end
  hue = hue * 60
  if hue < 0 then hue = hue + 360 end
  return hue
end
D.colToHue = colToHue

local function getItemID(val)
  local itemString = select(2, GetItemInfo(val))
  if type(itemString) == "string" then
    local subString = itemString:sub(itemString:find(":") + 1)
    subString = subString:sub(1, subString:find(":") - 1)
    return tonumber(subString)
  end
end
D.getItemID = getItemID



-------------------
-- Object-oriented
-------------------


local function setIndex(obj, ind, noOverride)
  local meta = getmetatable(obj)
  if meta == nil then
    meta = {}
    setmetatable(obj, meta)
  end
  if noOverride and meta.__index then return end
  meta.__index = ind
end
D.setIndex = setIndex

local Inherits = {}
function Inherits:New()
  self.__index = self
  local new = {}
  setmetatable(new, self)
  new._super = self
  return new
end

function Inherits:ForAll(funcName, before, subArgs, after)
  local hasBefore, hasAfter = not empty(before), not empty(after)
  if hasBefore and hasAfter then
    for _,arg in ipairs(subArgs) do self[funcName](self, unpack(before), arg, unpack(after)) end
  elseif hasBefore then
    for _,arg in ipairs(subArgs) do self[funcName](self, unpack(before), arg) end
  elseif hasAfter then
    for _,arg in ipairs(subArgs) do self[funcName](self, arg, unpack(after)) end
  else
    for _,arg in ipairs(subArgs) do self[funcName](self, after) end
  end
end


------------
-- Database
------------

local Key = Inherits:New()
Key:New()
function Key:New(copyVal)
  copyVal = copyVal or {}
  assertType(copyVal,"table", 3)
  local val = { unpack(copyVal) }
  setmetatable(val, self)
  return val
end
function Key:s(size) -- Sub-array
  assertType(size,"number", 3)
  local newVal = Key:New{}
  for i=1,size do
    newVal[i] = self[i]
  end
  return newVal
end
function Key:r(remove) -- Remove
  assertType(remove,"number", 3)
  return self:s(#self-remove)
end
function Key:a(...) -- Append
  local t_insert = table.insert
  local newVal = Key:New(self)
  for _,v in ipairs{...} do
    t_insert(newVal, v)
  end
  return newVal
end
D.Key = Key

local function K(val)
  return Key:New(val)
end
D.K = K

local function npairs(val, depth, ...)
  if val == nil or type(val) ~= "table" then return pairs{} end
  local i = {}
  local done = false

  if depth <= 1 then return function()
    if done then return end
    done = true
    return val
  end end

  local subPairs = npairs(val, depth - 1, select(2, ...))
  local result = {}
  local check = ...

  local function assignNext()
    local subVal = result[depth - 1]
    if subVal == nil or type(subVal) ~= "table" then
      i[depth -1], i[depth] = nil, nil
      return
    end
    while true do
      i[depth -1], i[depth] = next(subVal, i[depth -1])
      if i[depth -1]==nil then return end
      local checkType = type(check)
      if not check and i[depth -1]~=nil then return end
      if checkType == "function" and check(i[depth]) then return end
    end
  end

  return function()
    assignNext()
    while not done and i[depth] == nil do
      result = { subPairs() }
      if result[1] == nil then
        done = true
        return
      else
        i = Key:New(result):r(1)
        assignNext()
      end
    end
    if not done then return unpack(i) end
  end
end
D.npairs = npairs


-------------------
-- Crawling
-------------------

local function crawl(obj, key, noChildren, raw)
  assertType(obj,"table", key,"table", 3)
  local max = #key
  if max == 0 then return obj elseif max < 0 then return nil end
  local node = obj
  local child
  for i=1,max do
    if raw then
      child = rawget(node, key[i])
    else
      child = node[key[i]]
    end
    if not noChildren and child == nil then
      child = {}
      node[key[i]] = child
    end
    node = child
    if node == nil then return end
  end
  return node
end
D.crawl = crawl

local NodeCrawler = Inherits:New()
function NodeCrawler:New(core)
  core = core or {}
  assertType(core,"table", 3)
  local new = NodeCrawler._super.New(self)
  new.core = core
  return new
end
function NodeCrawler:Get(key)
  assertType(key,"table", 3)
  return crawl(self.core, key, true)
end
function NodeCrawler:Set(key, value, noOverride)
  assertType(key,"table", 3)
  if empty(key) then
    if value == nil then value = {} end
    if not noOverride then self.core = value end
    return self.core
  end
  local parent = crawl(self.core, K(key):r(1), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local child = key[#key]
  if not (noOverride and parent[child] ~= nil) then
    parent[child] = value
  end
  return parent[child]
end
function NodeCrawler:GetOrMake(key)
  assertType(key,"table", 3)
  return self:Set(key, {}, true)
end
function NodeCrawler:Incr(key, amount)
  amount = amount or 1
  assertType(key,"table", amount,"number", 3)
  local parent = crawl(self.core, K(key):r(1), false, true)
  local child = key[#key]
  parent[child] = amount + (parent[child] or 0)
  return parent[child]
end
function NodeCrawler:Append(key, ...)
  assertType(key,"table", 3)
  local t_insert = table.insert
  local parent = crawl(self.core, K(key):r(1), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local childKey = key[#key]
  local child = parent[childKey]
  if child == nil then
    child = {}
    parent[childKey] = child
  elseif type(child) ~= "table" then
    return
  end
  for _,el in ipairs{...} do
    t_insert(parent[childKey], el)
  end
end
D.NodeCrawler = NodeCrawler

local Configuration = NodeCrawler:New()
function Configuration:New(val)
  local new = Configuration._super.New(self, val)
  return new
end
function Configuration:AutoDefault(child)
  if not child or not child.type or not self.defaults then return end
  local default = self.defaults[child.type]
  if default then setIndex(child, default, true) end
end

function Configuration:CrawlDefaults(node, crawled)
  crawled = crawled or {}
  if not node then node = self.core end
  crawled[node] = true
  if type(node) ~= "table" then return end
  self:AutoDefault(node)
  for k,v in pairs(node) do
    if not crawled[v] then
      if type(v) == "table" and not (type(k) == "string" and k:find("_")) then
        self:CrawlDefaults(v, crawled)
      end
    end
  end
end
function Configuration:Make(parent, makeType, childName, noOverride)
  assertType(parent,"table", makeType,"string", childName,{"string","number"}, 3)
  if type(childName) == "string" then childName = childName:gsub("_","") end
  if empty(childName) then return end
  if noOverride and Configuration:Overriden(parent, childName) then return parent[childName] end
  local child = parent[childName]
  if child == nil then
    child = {}
    parent[childName] = child
  end
  child.type = makeType
  self:AutoDefault(child)
  return child
end
function Configuration:GetAll(parent, getType, depth)
  if empty(parent) then return pairs{} end
  depth = depth or 2
  assertType(parent,"table", getType,"string", depth,"number", 3)
  return npairs(parent, depth,
    function(val)
      return type(val) == "table" and val.type == getType
    end
  )
end
function Configuration:Count(parent, getType, depth)
  assertType(parent,"table", getType,"string", 3)
  local count = 0
  for ele in self:GetAll(parent, getType, depth) do
    count = count+1
  end
  return count
end
function Configuration:Delete(parent, childName)
  assertType(parent,"table", childName,{"string","number"}, 3)
  parent[childName] = nil
  parent._exclude = parent._exclude or {}
  parent._exclude[childName] = true
end
function Configuration:Excluded(parent, childName)
  assertType(parent,"table", childName,{"string","number"}, 3)
  if type(parent) ~= "table" then return end
  local exclude = parent._exclude
  return exclude and exclude[childName]
end
function Configuration:Overriden(parent, childName)
  assertType(parent,"table", childName,{"string","number"}, 3)
  local child = parent[childName]
  if child == nil then return false end
  return child.type ~= nil
end
function Configuration:ConfGet(info)
  local val = self:Get(info)

  if type(val) == "table" and val.r and val.g and val.b then
    return val.r, val.g, val.b, val.a end
  return val
end
function Configuration:ConfGetString(info)
  return tostring(self:ConfGet(info))
end
function Configuration:ConfSet(info, r, g, b, a)
  if b then return self:Set(info, {r = r, g = g, b = b, a = a })
  elseif tonumber(r) then return self:Set(info, tonumber(r))
  else return self:Set(info, r) end
end
function Configuration:GetParent(info)
  local parent = self:Get(K(info):r(1))
  if type(parent) == "table" then return parent end
end
function Configuration:GetParentName(info)
  return info[#info - 1]
end
function Configuration:ConfGetAll(info)
  local parent = self:GetParent(info)
  if parent then return self:GetAll(parent, info[#info]) end
end
function Configuration:ConfDelete(info)
  local parent = self:GetParent(info)
  if parent then return self:Delete(parent, info[#info]) end
end
function Configuration:ConfExcluded(info)
  local parent = self:GetParent(info)
  if parent then return self:Excluded(parent, info[#info]) end
end
function Configuration:ConfOverriden(info)
  local parent = self:GetParent(info)
  if parent then return self:Overriden(parent, info[#info]) end
end
function Configuration:Number(info, val)
  return val and type(tonumber(val)) == "number"
end
function Configuration:Locked(info)
  local parent = self:GetParent(info)
  if parent then return parent.lock end
end

D.Configuration = Configuration


-----------------
-- Addon handler
-----------------

local Handler = Configuration:New()

function Handler:New(addon)
  assertType(addon,"table", 3)
  local handler = Handler._super.New(self)
  setmetatable(addon, Handler)
  local addon_settings = addon.settings
  if addon_settings then
    addon.defaults = addon_settings.defaults
    addon_settings.handler = addon
  end
  addon.timers = {}
  return addon
end
function D.Addon(shortName, name, ...)
  local addon = LibStub("AceAddon-3.0"):NewAddon(shortName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", ...)
  addon.shortName = shortName
  addon.name = name
  addon.settings = LibStub(shortName .. "-Settings")
  addon.exports = LibStub:NewLibrary(shortName .. "-Exports", 1)
  return Handler:New(addon)
end
function Handler:Output(...)
  if self.core._debug and self.core._debug.print then print(self.name,"|",...) end
end
function Handler:GetVersion()
  return self.core._version
end
function Handler:SetVersion(version)
  self.core._version = version
end
function Handler:ResetDB()
  self.lib.db:ResetProfile()
end
function Handler:Queue(funcName, ...)
  local timers = self.timers
  if type(self[funcName]) ~= "function" then error(say("Invalid function ", funcName,"."), 2) end
  if not self.ScheduleTimer then error("AceTimer not mixed in!", 2) end
  if not timers[funcName] or self:TimeLeft(timers[funcName]) == 0 then
    timers[funcName] = self:ScheduleTimer(funcName,0.01,...)
  end
end
function Handler:TrackEvent(funcName, event, ...)
  event = event or funcName
  self:RegisterEvent(event, funcName, ...)
end
function Handler:TrackEventQueue(funcName, event, ...)
  event = event or funcName
  self:RegisterEvent(event, "Queue", funcName, ...)
end
function Handler:TrackMultiEvent(funcName, ...)
  self:ForAll("TrackEvent", { funcName }, {...})
end
function Handler:TrackMultiEventQueue(funcName, ...)
  self:ForAll("TrackEventQueue", { funcName }, {...})
end
function Handler:Reset(registered)
  print(self.shortName,0)
  self.core = self.lib.db.profile
  self:CrawlDefaults()
  if registered ~= false then
    self:SetVersion(self.version)
  end
  if self:ConfGetDebug{ "export" } then
    _G[self.shortName .. "Debug"] = self
  else
    _G[self.shortName .. "Debug"] = nil
  end

  print(self.shortName,1)
  local settings = self.settings
  if settings then
    print(self.shortName,2)
    settings.handler = self
    settings.options = { type = "group", args = {} }
    local defaults = self.defaults
    if defaults and defaults.options then
      print(self.shortName,3)
      settings.options = defaults.options
    end
    settings.options.handler = self
    settings.crawler = NodeCrawler:New(settings.options.args)

    if defaults and defaults.global then
      print(self.shortName,4)
      setIndex(self.core, defaults.global)
    end
  end
  self:OnReset(registered)
  if self.OnLoad then self:OnLoad(registered) end

  if settings and settings.options then
    self.lib.registry:RegisterOptionsTable(self.name, settings.options)
  end
end

function Handler:MigrateDB(...)
  local keys = { ... }
  local old = Configuration:New{}
  for i,key in ipairs(keys) do
    key = K(key)
    if key == nil then key = {} end
    local subTable = self:Get(key)
    if type(subTable) ~= "table"then
      old:Set(key, subTable)
    else
      for k,v in pairs(subTable) do
        local subKey = key:a(k)
        if type(v) ~= "table" then old:Set(subKey, v) end
      end
    end
  end

  self:ResetDB()

  for i,key in ipairs(keys) do
    key = K(key)
    local subTable = old:Get(key)
    if type(subTable) ~= "table" then
      if subTable ~= self:ConfGet(key) then self:ConfSet(key, subTable) end
    else for k,v in pairs(subTable) do
      local subKey = key:a(k)
      if type(v) ~= "table" and v ~= self:ConfGet(subKey) then self:ConfSet(subKey, v) end
    end end
  end

  if self.version then self:SetVersion(self.version) end
end

local function debugKey(info)
  return K{ "_debug" }:a(unpack(info))
end
function Handler:ConfGetDebug(info)
  return self:ConfGet(debugKey(info))
end
function Handler:ConfSetDebug(info, r, g, b, a)
  self:ConfSet(debugKey(info), r, g, b, a)
end
function Handler:DebugExport(info, val)
  if val then
    _G[self.shortName .. "Debug"] = self
  else
    _G[self.shortName .. "Debug"] = nil
  end
  return self:ConfSetDebug(info,val)
end
function Handler:DebugOptions(args)
  local debug = {
    type = "group",
    handler = self,
    get = "ConfGetDebug",
    set = "ConfSetDebug",
    args = {
      export = {
        type = "toggle",
        name = "Export to global variable (" .. self.shortName .. "Debug)",
        width = "full",
        set = "DebugExport",
        order = 0
      },
      print = {
        type = "toggle",
        name = "Print debug messages",
        width = "full",
        set = "ConfSetDebug",
        order = 1
      },
    }
  }
  local debug_args = debug.args
  for k,v in pairs(args) do
    debug_args[k] = v
  end
  return debug
end


local dbCallbacks = { "OnProfileCopied", "OnProfileDeleted", "OnProfileReset", "OnDatabaseReset", "OnProfileChanged" }
function Handler:Register(libs,defaults,defaultProfile)
  self.lib = self.lib or {}
  local name, lib, settings = self.name, self.lib, self.settings
  if settings and settings.custom then
    settings.defaults = settings.defaults or {}
    local subDefaults = settings.defaults
    local battletag = select(2, BNGetInfo())
    for name, custom in pairs(settings.custom) do
      if name == battletag then
        for customType, customSettings in pairs(custom) do
          subDefaults[customType] = subDefaults[customType] or {}
          local subDefaultType = subDefaults[customType]
          for propertyName, propertyVal in pairs(customSettings) do
            subDefaultType[propertyName] = propertyVal
          end
        end
      end
    end
    if not defaults and subDefaults.db then
      defaults = settings.defaults.db
    end
  end
  defaults = defaults or {}

  debugSelf = self
  lib.db = LibStub("AceDB-3.0"):New(self.shortName .. "DB", defaults, defaultProfile)
  lib.registry = LibStub("AceConfigRegistry-3.0")
  lib.settings = LibStub("AceConfigDialog-3.0")
  lib.options = LibStub("AceDBOptions-3.0")
  for libKey, libName in pairs(libs or {}) do
    if not lib[libKey] then
      lib[libKey] = LibStub(libName,true)
    end
  end

  self:Reset(false)

  lib.settings:AddToBlizOptions(name)
  local profileSettings = LibStub("AceDBOptions-3.0"):GetOptionsTable(lib.db)
  lib.registry:RegisterOptionsTable(name .. " Profile", profileSettings)
  lib.settings:AddToBlizOptions(name .. " Profile", "Profiles", name)

  for _,callback in ipairs(dbCallbacks) do
    lib.db.RegisterCallback(self, callback, "Reset")
  end

  if settings and settings.debug then
    local debug = self:DebugOptions(settings.debug)
    lib.registry:RegisterOptionsTable(name .. " Extras", debug)
    lib.settings:AddToBlizOptions(name .. " Extras","Extras", name)
  end
end
D.Handler = Handler


-----------
-- Anchors
-----------

local anchors = {
  center      = "Center",
  topleft     = "Top Left",
  top         = "Top",
  topright    = "Top Right",
  left        = "Left",
  right       = "Right",
  bottomleft  = "Bottom Left",
  bottom      = "Bottom",
  bottomright = "Bottom Right"
}
D.anchors = anchors

local VERTICAL = 1
D.VERTICAL = VERTICAL
local HORIZONTAL = 2
D.HORIZONTAL = HORIZONTAL
local orientations = {
  horizontal = "Horizontal",
  vertical = "Vertical"
}
D.orientations = orientations
local orientGrowth = {
  horizontal = {
    [false] = "right",
    [true] = "left"
  },
  vertical = {
    [false] = "top",
    [true] = "bottom"
  }
}
D.orientGrowth = orientGrowth
local direction = {
  TOP    = VERTICAL,    top    = VERTICAL,
  BOTTOM = VERTICAL,    bottom = VERTICAL,
  LEFT   = HORIZONTAL,  left   = HORIZONTAL,
  RIGHT  = HORIZONTAL,  right  = HORIZONTAL,
  horizontal = HORIZONTAL,
  vertical = VERTICAL
}
D.direction = direction
local function compound(anchor1, anchor2)
  if not anchor2 then return anchor1 end
  if not anchor1 then return anchor2 end
  if direction[anchor1] < direction[anchor2] then
    return anchor1 .. anchor2
  else
    return anchor2 .. anchor1
  end
end
D.compound = compound

local grow = { left = "Left", right = "Right", top = "Up", bottom = "Down" }
D.grow = grow

local growAnchors = {
  left   = { "RIGHT",  "LEFT",  -1,   0 },
  right  = { "LEFT",   "RIGHT",  1,   0 },
  top    = { "BOTTOM", "TOP",    0,   1 },
  bottom = { "TOP",    "BOTTOM", 0,  -1 },
  blank  = { "CENTER", "CENTER", 0,   0 }
}
D.growAnchors = growAnchors
for _,dir1 in ipairs{ "left", "right" } do for _,dir2 in ipairs{ "top", "bottom" } do
  growAnchors[compound(growAnchors[dir1][1],growAnchors[dir2][1])] = { nil, nil, growAnchors[dir1][3], growAnchors[dir2][4] }
end end

local corners = {
  topleft     = { "BOTTOMRIGHT", "TOPLEFT",     -1, -1 },
  topright    = { "BOTTOMLEFT",  "TOPRIGHT",     1, -1 },
  bottomleft  = { "TOPRIGHT",    "BOTTOMLEFT",  -1,  1 },
  bottomright = { "TOPLEFT",     "BOTTOMRIGHT",  1,  1 }
}
D.corners = corners

local borders = {
  top    = { { "BOTTOMLEFT", "TOPLEFT",    -1,  0 }, { "BOTTOMRIGHT", "TOPRIGHT",     1,  0 },  0,  1 },
  bottom = { { "TOPLEFT",    "BOTTOMLEFT", -1,  0 }, { "TOPRIGHT",    "BOTTOMRIGHT",  1,  0 },  0,  1 },
  left   = { { "TOPRIGHT",   "TOPLEFT",     0,  1 }, { "BOTTOMRIGHT", "BOTTOMLEFT",   0, -1 },  1,  0 },
  right  = { { "TOPLEFT",    "TOPRIGHT",    0,  1 }, { "BOTTOMLEFT",  "BOTTOMRIGHT",  0, -1 },  1,  0 }
}
D.borders = borders

local function convertDims(dir, dim1, dim2)
  if direction[dir] == HORIZONTAL then
    return dim1, dim2
  else
    return dim2, dim1
  end
end
D.convertDims = convertDims

local function makeGrid(conf)
  local math_min, math_max, math_ceil = math.min, math.max, math.ceil
  conf.padding = conf.padding or 0
  conf.spacing = conf.spacing or 0

  if not conf.grow2 then
    if direction[conf.grow] == HORIZONTAL then
      conf.grow2 = conf.rowGrowth
    else
      conf.grow2 = conf.columnGrowth
    end
  end

  local conf_frame, conf_parent = conf.frame, conf.parent
  local conf_els, conf_max, conf_limit = conf.els, conf.max, conf.limit
  local conf_anchor, conf_grow, conf_grow2 = conf.anchor, conf.grow, conf.grow2
  local conf_size, conf_spacing, conf_padding = conf.size, conf.spacing, conf.padding
  local conf_x, conf_y = conf.x, conf.y
  local conf_background, conf_border, conf_edge = conf.background, conf.border, conf.edge

  local a1 = growAnchors[conf_grow] or growAnchors.blank
  local a2 = growAnchors[conf_grow2] or growAnchors.blank
  local comp = compound(a1[1], a2[1])
  local growComp = growAnchors[comp]

  conf_frame:ClearAllPoints()
  conf_frame:SetPoint(comp, conf_parent, conf_anchor,
    conf_x - (conf_size + conf_padding * 2) * growComp[3] / 2,
    conf_y - (conf_size + conf_padding * 2) * growComp[4] / 2)

  if not conf_els or #conf_els == 0 then
    conf_frame:SetSize(conf_size + conf_padding * 2, conf_size + conf_padding * 2)
    return
  end

  conf_els[1]:ClearAllPoints()
  conf_els[1]:SetPoint(comp, conf_frame, comp, growComp[3] * conf_padding, growComp[4] * conf_padding)

  local max = #conf_els
  if conf_max and conf_max > 0 then max = math_min(max, conf_max) end

  for i=2,max do
    conf_els[i]:ClearAllPoints()

    if conf_limit and conf_limit > 0 and (i-1) % conf_limit == 0 then
      conf_els[i]:SetPoint(a2[1], conf_els[i-conf_limit], a2[2], a2[3] * conf_spacing, a2[4] * conf_spacing)
    else
      conf_els[i]:SetPoint(a1[1], conf_els[i-1], a1[2], a1[3] * conf_spacing, a1[4] * conf_spacing)
    end
  end
  local dim1, dim2
  if conf_limit and conf_limit > 0 and max > conf_limit then
    dim1 = conf_padding * 2 + conf_size * conf_limit + conf_spacing * (conf_limit - 1)
    local subs = math_ceil(max/conf_limit)
    dim2 = conf_padding * 2 + conf_size * subs + conf_spacing * (subs - 1)
  else
    dim1 = conf_padding * 2 + conf_size * max + conf_spacing * (max - 1)
    dim2 = conf_padding * 2 + conf_size
  end
  conf_frame:SetSize(convertDims(conf_grow, dim1, dim2))
  local bg = {}
  if conf_background and conf_background ~= "None" then
    bg.bgFile = SharedMedia:Fetch(SharedMedia.MediaType.BACKGROUND, conf.background)
  end
  if conf_border and conf_border ~= "None" then
    bg.edgeFile = SharedMedia:Fetch(SharedMedia.MediaType.BORDER, conf.border)
    bg.edgeSize = conf_edge or 4
  end
  conf_frame:SetBackdrop(bg)
  conf_frame:Show()
end
D.makeGrid = makeGrid